/*

Data structures for loading and storing PLY models

Created by Eugene Zhang 2005

Modified by Guoning Chen 2012

*/

//#pragma once

#ifndef __SKELETON_H__
#define __SKELETON_H__

#pragma once

//#include "ply.h"
#include "icVector.H"
//#include "icMatrix.h"
#include <stdio.h>

#include <vector>
using namespace std;

const double EPS = 1.0e-6;
const double PI=3.1415926535898;

/* forward declarations */
class Triangle;
class Edge;

/** Corner **/
class Corner{
public:
	unsigned char Edge_count;   //special variable for edges search 1/21/05
	int index;
	int v;            //the ID of the vertex of the corner
	int n;            //the next corner according to the orientation
	int p;            //the previous corner according to the orientation
	int t;            //the triangle the corner belongs to
	int ot;           //the index of its opposite triangle for traversal 2/9/05
	int o;            //the index of its opposite corner
	Edge *e;          //the opposite edge of the corner
	float angle;       //the angle of the corner
	float BeginAng, EndAng;  //for correct angle allocation of the corner around vertex v
	float r;
	bool orient;
	
	/////////////////////////////////////////////////
	/* Optional variables */
	Edge *edge[2];    //two edges associated with this corner

	Corner()
	{
		e = NULL;
		edge[0] = edge[1] = NULL;
	}

	double get_Angle();
};


class CornerList{
public:
	Corner **corners;
	int ncorners;
	int curMaxNumCorners;

	// The similar list operations as VertexList
	CornerList(int initsize = 1000) //construction
	{
		corners = (Corner **)malloc(sizeof(Corner *)*initsize);
		curMaxNumCorners = initsize;
		ncorners = 0;
	} 

	~CornerList()
	{
		if(corners == NULL)
			return;

		int i;

		for(i = 0; i < ncorners; i++)
		{
			free(corners[i]);
		}
		free(corners);
		corners = NULL;
	}

	inline void finalize()
	{
		if(corners != NULL)
		{
			int i;

			for(i = 0; i < ncorners; i++)
			{
				free(corners[i]);
			}
			free(corners);

			corners = NULL;
			ncorners = 0;
			curMaxNumCorners = 0;
		}
	}

	//add a new vertex to the end of the list, if it succeeds, return true
	inline bool append(Corner *c)
	{
		if(isFull ())
			if(!extend(100))
				return false;             //if not enough memory available, return false
		corners[ncorners] = c;

		ncorners++;
		return true;
	} 

	inline bool del_End() //delete the vertex at the end of the list
	{
		if(isEmpty())  return false;
		ncorners --;
		return true;
	} 

	inline void copy_Elem(Corner *s, Corner *d)
	{
	}

	//delete the corresponding  vertex, if it succeeds, return true
	inline bool del_Node(Corner *c) 
	{
		if(isEmpty())  return false;

		//find the vertex, if find it, delete and move the following vertices forward
		//otherwise, return false;

		int i, pos = -1;

		for(i = 0; i < ncorners; i++)
		{
			if(corners[i] == c)
			{
				pos = i;
				break;
			}

		}

		if(pos == -1) return false;

		//delete it
		for(i = pos; i < ncorners-1; i++)
		{
			//we need a copy function
			copy_Elem(corners[i], corners[i+1]);
		}

		ncorners--;

		return true;
	} 

	inline bool isEmpty()  //judge whether the list is empty
	{
		if(ncorners == 0)   return true;
		return false;
	}

	inline bool isFull()
	{
		if(ncorners == curMaxNumCorners) return true;
		return false;
	}

	//extend the original list, if it succeeds, return true
	inline bool extend(int step = 100)
	{
		Corner **temp = corners;
		corners = (Corner **) malloc(sizeof(Corner *) * (curMaxNumCorners + step));
		if( corners == NULL)
		{
			//fail
			corners = temp;
			return false;
		}

		int i;

		for(i = 0; i < curMaxNumCorners; i++)
			corners[i] = temp[i];
		
		for(i = curMaxNumCorners; i < curMaxNumCorners+step; i++)
			corners[i] = NULL;

		curMaxNumCorners += step;
		free(temp);
		return true;
	}

	inline void reset()
	{
		ncorners = 0;
	}


	//void build_Corners();    //build the corner table

}; // end of CornerList class


class Vertex {
public:
	double x,y,z;
	double vx, vy, vz;
	int index;

	int ntris;
	Triangle **tris;
	unsigned char max_tris;

	Edge **adjedges;    //store the addresses of all adjacent edges
	unsigned char num_adjedges;

	icVector3 normal;
	Corner **corners;       //the corners sharing the vertex
	unsigned char ncorners;

    // For curvature estimation
	icVector3 T, B;
	float MaxCurvature;       //store the maximum curvature
	float MinCurvature;       //store the minimum curvature 
	icVector3 MaxDirection;   //store the maximum curvature direction
	icVector3 MinDirection;   //store the minimum curvature direction

	float MeanCurvature;
	float GaussianCurvature;

	// vector field defined on the vertex
	icVector2 vec;

	double *Anglist;   //store the angle allocated for each triangle on the tangent plane

	void *other_props;

public:
  Vertex(double xx, double yy, double zz) { x = xx; y = yy; z = zz; }
};

class Edge {
public:
  int index;
  Vertex *verts[2];
  int ntris;
  Triangle **tris;
	double length;
	
	float NormalCurvature0; //store the normal curvature along the edge
	float NormalCurvature1; //store the normal curvature along the edge
};

class Triangle {
public:
  int index;
  int nverts;
  Vertex *verts[3];
  Edge *edges[3];

	double angle[3];
	float area;

	double x1, x2, y2;

	icVector3 normal, LX, LY;
	icVector2 loc_vec[3];         //the local vector values at the three vertices

	void *other_props;

  void compute_local_frame();
  void get_barycentric_coordinates(double gp[3], double alpha[3]);
  void get_barycentric_coordinates_loc(double lp[2], double alpha[3]);
  icVector2 get_vector_at_pt(double lp[2]); 
	bool fall_in_the_tri(double pos[3]);

};

class VertexList {

public:

  int num_verts;
  int max_verts;
  Vertex **verts;

  VertexList (int max) {
    max_verts = max;
    verts = new Vertex *[max_verts];
    num_verts = 0;
  }
	void finalize() {
		if (this != NULL) {
			free(verts);
			free(this);
		}
	}
	void append(Vertex *v)
	{
		int i;

		/* first make sure there is enough room for new vertex */

		if (num_verts >= max_verts) {
			max_verts += 10;
			Vertex **tlist = new Vertex *[max_verts];
			for (i = 0; i < num_verts; i++)
				tlist[i] = verts[i];
			delete (verts);
			verts = tlist;
		}

		/* add new vertex to list */

		verts[num_verts] = v;
		num_verts++;
	}

};

class TriangleList {
public:

  int num_tris;
  int max_tris;
  Triangle **tris;

  TriangleList (int max) {
    max_tris = max;
    tris = new Triangle *[max_tris];
    num_tris = 0;
  }
	void append(Triangle *t)
	{
		int i;

		/* first make sure there is enough room for new triangle */

		if (num_tris >= max_tris) {
			max_tris += 10;
			Triangle **tlist = new Triangle *[max_tris];
			for (i = 0; i < num_tris; i++)
				tlist[i] = tris[i];
			delete (tris);
			tris = tlist;
		}

		/* add new triangle to list */

		tris[num_tris] = t;
		num_tris++;
	}
};

class EdgeList {

public:

  int num_edges;
  int max_edges;
  Edge **edges;

  EdgeList (int max) {
    max_edges = max;
    edges = new Edge *[max_edges];
    num_edges = 0;
  }
	void append(Edge *e)
	{
		int i;

		/* first make sure there is enough room for new edge */

		if (num_edges >= max_edges) {
			max_edges += 10;
			Edge **tlist = new Edge *[max_edges];
			for (i = 0; i < num_edges; i++)
				tlist[i] = edges[i];
			delete (edges);
			edges = tlist;
		}

		/* add new edge to list */

		edges[num_edges] = e;
		num_edges++;
	}
};

class Polyhedron {
public:

	int index;

	Triangle **tlist;  /* list of triangles */
	int ntris;
	int max_tris;

	Vertex **vlist;    /* list of vertices */
	int nverts;
	int max_verts;

	Edge **elist;      /* list of edges */
	int nedges;
	int max_edges;

	icVector3 center;
	double radius;
	double area;

	int seed;
	CornerList clist;          //the corner list


  //PlyOtherProp *vert_other,*face_other;
	
	Polyhedron();
    Polyhedron(FILE *);

	//// Assume you have implemented the Corner table corrected in the first assignment
	//// what you need for this assignment is to sort the Corners on each vertex consistently

	void build_Corners();  //build the corner table
	void add_Corner_to_Vertex(Corner *, Vertex *v);
	void find_Opposite();
	void sort_Corner_on_Vertex();
	void alloc_Corner_Angle();
	int get_orientation(Vertex *p, double cur_ang, Edge *e1);

	void average_normals();

	void create_edge(Vertex *, Vertex *);
	void create_edges();
	int face_to_vertex_ref(Triangle *, Vertex *);
	void order_vertex_to_tri_ptrs(Vertex *);
	void vertex_to_tri_ptrs();
	Triangle *find_common_edge(Triangle *, Vertex *, Vertex *);
	Triangle *other_triangle(Edge *, Triangle *);
	void calc_bounding_sphere();
	void calc_face_normals_and_area();
	void calc_edge_length();

	void write_file(FILE *);

	void create_pointers();

	// initialization and finalization
	void initialize();
	void finalize();


	// For the oriented bounding box (Project 1)
	icVector3 max_dir, mid_dir, min_dir;
	double max_ev, mid_ev, min_ev;
	icVector3 mass_center;

	void  vertex_to_edge_ptrs();

};


#endif /*__SKELETON_H__*/